"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_ux_1 = tslib_1.__importDefault(require("cli-ux"));
const debug_1 = tslib_1.__importDefault(require("debug"));
const fsExtra = tslib_1.__importStar(require("fs-extra"));
const path = tslib_1.__importStar(require("path"));
const pg_query_stream_1 = tslib_1.__importDefault(require("pg-query-stream"));
const lisk_genesis_1 = require("@liskhq/lisk-genesis");
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_utils_1 = require("@liskhq/lisk-utils");
const lisk_tree_1 = require("@liskhq/lisk-tree");
const storage_1 = require("./storage");
const schema_1 = require("./schema");
const debug = debug_1.default('lisk_migrator:genesis_block');
const sortBufferArray = (arr) => arr.sort((a, b) => a.compare(b));
const SIZE_INT32 = 4;
const SIZE_INT64 = 8;
exports.getBlockBytes = (block) => {
    var _a;
    const blockVersionBuffer = lisk_cryptography_1.intToBuffer(block.version, SIZE_INT32, lisk_cryptography_1.LITTLE_ENDIAN);
    const timestampBuffer = lisk_cryptography_1.intToBuffer(block.timestamp, SIZE_INT32, lisk_cryptography_1.LITTLE_ENDIAN);
    const previousBlockBuffer = block.previousBlock
        ? lisk_cryptography_1.intToBuffer(block.previousBlock, SIZE_INT64, lisk_cryptography_1.BIG_ENDIAN)
        : Buffer.alloc(SIZE_INT64);
    const numTransactionsBuffer = lisk_cryptography_1.intToBuffer(block.numberOfTransactions, SIZE_INT32, lisk_cryptography_1.LITTLE_ENDIAN);
    const totalAmountBuffer = lisk_cryptography_1.intToBuffer(block.totalAmount.toString(), SIZE_INT64, lisk_cryptography_1.LITTLE_ENDIAN);
    const totalFeeBuffer = lisk_cryptography_1.intToBuffer(block.totalFee.toString(), SIZE_INT64, lisk_cryptography_1.LITTLE_ENDIAN);
    const rewardBuffer = lisk_cryptography_1.intToBuffer(block.reward.toString(), SIZE_INT64, lisk_cryptography_1.LITTLE_ENDIAN);
    const payloadLengthBuffer = lisk_cryptography_1.intToBuffer(block.payloadLength, SIZE_INT32, lisk_cryptography_1.LITTLE_ENDIAN);
    const blockSignatureBuffer = (_a = block.blockSignature) !== null && _a !== void 0 ? _a : Buffer.alloc(0);
    return Buffer.concat([
        blockVersionBuffer,
        timestampBuffer,
        previousBlockBuffer,
        numTransactionsBuffer,
        totalAmountBuffer,
        totalFeeBuffer,
        rewardBuffer,
        payloadLengthBuffer,
        block.payloadHash,
        block.generatorPublicKey,
        blockSignatureBuffer,
    ]);
};
exports.accountDefaultProps = {
    token: {
        balance: BigInt(0),
    },
    sequence: {
        nonce: BigInt(0),
    },
    dpos: {
        delegate: {
            lastForgedHeight: 0,
            username: '',
            pomHeights: [],
            consecutiveMissedBlocks: 0,
            isBanned: false,
            totalVotesReceived: BigInt(0),
        },
        sentVotes: [],
        unlocking: [],
    },
    keys: {
        mandatoryKeys: [],
        optionalKeys: [],
        numberOfSignatures: 0,
    },
};
const MAX_EIGHT_BYTE_UNSIGNED_INTEGER = BigInt(2) ** BigInt(64) - BigInt(1);
exports.migrateAddressForAccount = (legacyAccount) => {
    const address = BigInt(legacyAccount.address.slice(0, -1));
    const eightByteAddress = Buffer.alloc(8);
    if (address > MAX_EIGHT_BYTE_UNSIGNED_INTEGER) {
        eightByteAddress.writeBigUInt64BE(address >> BigInt(64));
    }
    else {
        eightByteAddress.writeBigUInt64BE(address);
    }
    const twentyByteAddress = legacyAccount.publicKey && legacyAccount.outgoingTxCount > 0
        ? lisk_cryptography_1.getAddressFromPublicKey(legacyAccount.publicKey)
        : null;
    return { legacyAddress: eightByteAddress, newAddress: twentyByteAddress !== null && twentyByteAddress !== void 0 ? twentyByteAddress : eightByteAddress };
};
exports.migrateLegacyAccount = async ({ db, legacyAccount, snapshotHeight, accountsMap, delegatesMap, legacyAddressMap, }) => {
    debug('Migrating legacy account: ', legacyAccount.address);
    debug({
        incomingTxCount: legacyAccount.incomingTxCount,
        outgoingTxCount: legacyAccount.outgoingTxCount,
        balance: legacyAccount.balance,
    });
    if (legacyAccount.incomingTxCount === 0) {
        debug('No incoming transaction. Skipping.');
        return;
    }
    if (BigInt(legacyAccount.balance) < BigInt(0)) {
        debug('Its a genesis account. Skipping.');
        return;
    }
    const { legacyAddress, newAddress } = exports.migrateAddressForAccount(legacyAccount);
    debug('New address', newAddress.toString('hex'));
    if (accountsMap.has(newAddress) && newAddress.length === 20) {
        throw new Error(`Account with publicKey: ${legacyAccount.publicKey.toString('hex')} already migrated.`);
    }
    let duplicateAccount;
    if (accountsMap.has(newAddress)) {
        duplicateAccount = accountsMap.get(newAddress);
    }
    else if (legacyAddressMap.has(legacyAddress)) {
        const duplicateAddress = legacyAddressMap.get(legacyAddress);
        duplicateAccount = accountsMap.get(duplicateAddress);
    }
    if (duplicateAccount) {
        debug('Duplicate account detected. Adding up the balance.');
        const duplicateUpdatedAccount = {
            ...duplicateAccount,
            token: { balance: duplicateAccount.token.balance + BigInt(legacyAccount.balance) },
        };
        accountsMap.set(duplicateAccount.address, duplicateUpdatedAccount);
        return;
    }
    const basicProps = { address: newAddress };
    const tokenProps = { token: { balance: BigInt(legacyAccount.balance) } };
    const dposProps = {
        dpos: {
            delegate: {
                username: legacyAccount.isDelegate ? legacyAccount.username : '',
                lastForgedHeight: snapshotHeight + 1,
            },
        },
    };
    let keysProps = {};
    if (legacyAccount.secondSignature && legacyAccount.multimin > 0) {
        const publicKey = legacyAccount.publicKey;
        const keys = await db.manyOrNone(storage_1.SQLs.getMultisigPublicKeys, {
            address: legacyAccount.address,
        });
        keysProps = {
            keys: {
                mandatoryKeys: sortBufferArray([publicKey, legacyAccount.secondPublicKey]),
                optionalKeys: sortBufferArray(keys.map(r => r.publicKey)),
                numberOfSignatures: legacyAccount.multimin + 1,
            },
        };
    }
    else if (legacyAccount.secondSignature && legacyAccount.multimin === 0) {
        const publicKey = legacyAccount.publicKey;
        if (!publicKey.equals(legacyAccount.secondPublicKey)) {
            keysProps = {
                keys: {
                    mandatoryKeys: sortBufferArray([publicKey, legacyAccount.secondPublicKey]),
                    optionalKeys: [],
                    numberOfSignatures: 2,
                },
            };
        }
    }
    else if (!legacyAccount.secondSignature && legacyAccount.multimin > 0) {
        const keys = await db.manyOrNone(storage_1.SQLs.getMultisigPublicKeys, {
            address: legacyAccount.address,
        });
        keysProps = {
            keys: {
                optionalKeys: sortBufferArray(keys.map(r => r.publicKey)),
                mandatoryKeys: [legacyAccount.publicKey],
                numberOfSignatures: legacyAccount.multimin,
            },
        };
    }
    let account = lisk_utils_1.objects.mergeDeep({}, exports.accountDefaultProps, basicProps, tokenProps, keysProps);
    if (legacyAccount.username !== '') {
        account = lisk_utils_1.objects.mergeDeep({}, account, dposProps);
    }
    debug('New account:', account);
    legacyAddressMap.set(legacyAddress, account.address);
    accountsMap.set(account.address, account);
    if (account.dpos.delegate.username !== '') {
        delegatesMap.set(account.address, {
            address: account.address,
            votes: BigInt(legacyAccount.vote),
        });
    }
};
exports.sortByVotesReceived = (a, b) => {
    if (a.votes > b.votes) {
        return -1;
    }
    if (a.votes < b.votes) {
        return 1;
    }
    return a.address.compare(b.address);
};
exports.sortAccounts = (a, b) => {
    if (a.address.length < b.address.length) {
        return -1;
    }
    if (a.address.length > b.address.length) {
        return 1;
    }
    return a.address.compare(b.address);
};
exports.createGenesisBlockFromStorage = async ({ db, snapshotHeight, epochTime, }) => {
    const blockIDSubTreeRoots = [];
    const blocksBatchSize = 2 ** 16;
    const accountsBatchSize = 10000;
    let lastBlock;
    cli_ux_1.default.action.start('Processing blocks to calculate previous block id');
    const time = Date.now();
    const blocksStreamParser = (_, blocksBatch) => {
        blockIDSubTreeRoots.push(new lisk_tree_1.MerkleTree(blocksBatch.map(block => lisk_cryptography_1.hash(exports.getBlockBytes(block)))).root);
        lastBlock = blocksBatch[blocksBatch.length - 1];
        debug(`Processed block till height: ${lastBlock.height}, Remaining: ${snapshotHeight - lastBlock.height}`);
    };
    await db.stream(new pg_query_stream_1.default('SELECT * FROM blocks WHERE height <= $1 ORDER BY height ASC', [snapshotHeight], { batchSize: blocksBatchSize }), async (s) => storage_1.streamRead(s, blocksStreamParser));
    const merkleRootOfBlocksTillSnapshotHeight = new lisk_tree_1.MerkleTree(blockIDSubTreeRoots, {
        preHashedLeaf: true,
    }).root;
    cli_ux_1.default.action.stop(`done in ${Date.now() - time}ms`);
    const progress = cli_ux_1.default.progress({
        format: 'Migrating accounts: [{bar}] {percentage}% | {value}/{total} | ETA: {eta}',
        fps: 2,
        synchronousUpdate: false,
        etaAsynchronousUpdate: false,
        barsize: 50,
        stream: process.stdout,
        stopOnComplete: true,
    });
    const accountsMap = new lisk_utils_1.dataStructures.BufferMap();
    const delegatesMap = new lisk_utils_1.dataStructures.BufferMap();
    const legacyAddressMap = new lisk_utils_1.dataStructures.BufferMap();
    let processedAccounts = 0;
    const { totalAccounts } = await db.one('SELECT count(*) as "totalAccounts" FROM mem_accounts_snapshot');
    progress.start(totalAccounts, 0);
    const accountsStreamParser = async (_, data) => {
        for (const legacyAccount of data) {
            await exports.migrateLegacyAccount({
                db,
                legacyAccount,
                snapshotHeight,
                accountsMap,
                delegatesMap,
                legacyAddressMap,
            });
            processedAccounts += 1;
            progress.update(processedAccounts);
        }
    };
    await db.stream(new pg_query_stream_1.default('SELECT mem_accounts_snapshot.*, (SELECT COUNT(*)::int FROM trs WHERE trs."recipientId" = mem_accounts_snapshot.address) as "incomingTxCount", (SELECT COUNT(*)::int FROM trs WHERE trs."senderId" = mem_accounts_snapshot.address) as "outgoingTxCount" FROM mem_accounts_snapshot ORDER BY mem_accounts_snapshot."publicKey" ASC', [], { batchSize: accountsBatchSize }), async (s) => storage_1.streamRead(s, accountsStreamParser));
    await new Promise(resolve => {
        progress.on('stop', () => {
            resolve();
        });
    });
    const accounts = accountsMap.values().sort(exports.sortAccounts);
    const topDelegates = delegatesMap
        .values()
        .sort(exports.sortByVotesReceived)
        .slice(0, 103)
        .map(a => a.address);
    const epochTimeInMs = new Date(epochTime).getTime();
    const lastBlockTimeOffsetInMs = lastBlock.timestamp * 1000;
    const lastBlockTimeInSeconds = (epochTimeInMs + lastBlockTimeOffsetInMs) / 1000;
    const lastBlockTimeToNearest10s = Math.round(lastBlockTimeInSeconds / 10) * 10;
    cli_ux_1.default.action.start('Creating genesis block');
    const genesisBlock = lisk_genesis_1.createGenesisBlock({
        accounts: accounts,
        initRounds: 600,
        initDelegates: topDelegates,
        previousBlockID: merkleRootOfBlocksTillSnapshotHeight,
        height: snapshotHeight + 1,
        timestamp: lastBlockTimeToNearest10s + 7200,
        accountAssetSchemas: schema_1.defaultAccountSchema,
    });
    cli_ux_1.default.action.stop();
    return lisk_genesis_1.getGenesisBlockJSON({ genesisBlock, accountAssetSchemas: schema_1.defaultAccountSchema });
};
exports.writeGenesisBlock = (genesisBlock, filePath) => {
    if (fsExtra.existsSync(filePath)) {
        fsExtra.unlinkSync(filePath);
    }
    const { dir } = path.parse(filePath);
    fsExtra.ensureDirSync(dir);
    fsExtra.writeFileSync(filePath, JSON.stringify(genesisBlock, null, '\t'));
};
//# sourceMappingURL=genesis_block.js.map